@@ -1,8 +1,12 @@
+import errno
 import os
+import signal
 import struct
 import sys
 import termios
+import threading
 import types
+
 from io import BytesIO
 from itertools import chain, repeat
 
@@ -13,18 +17,22 @@ from pytest_relaxed import trap
 from mock import patch, Mock, call
 
 from invoke import (
-    Runner,
-    Local,
-    Context,
+    CommandTimedOut,
     Config,
+    Context,
     Failure,
-    ThreadException,
+    Local,
+    Promise,
     Responder,
-    WatcherError,
-    UnexpectedExit,
-    StreamWatcher,
     Result,
+    Runner,
+    StreamWatcher,
+    SubprocessPipeError,
+    ThreadException,
+    UnexpectedExit,
+    WatcherError,
 )
+from invoke.runners import default_encoding
 from invoke.terminals import WINDOWS
 
 from _util import (
@@ -43,6 +51,15 @@ class _RaisingWatcher(StreamWatcher):
         raise WatcherError("meh")
 
 
+class _GenericException(Exception):
+    pass
+
+
+class _GenericExceptingRunner(_Dummy):
+    def wait(self):
+        raise _GenericException
+
+
 def _run(*args, **kwargs):
     klass = kwargs.pop("klass", _Dummy)
     settings = kwargs.pop("settings", {})
@@ -99,7 +116,15 @@ def make_tcattrs(cc_is_ints=True, echo=False):
     return attrs
 
 
+class _TimingOutRunner(_Dummy):
+    @property
+    def timed_out(self):
+        return True
+
+
 class Runner_:
+    _stop_methods = ["generate_result", "stop", "stop_timer"]
+
     # NOTE: these copies of _run and _runner form the base case of "test Runner
     # subclasses via self._run/_runner helpers" functionality. See how e.g.
     # Local_ uses the same approach but bakes in the dummy class used.
@@ -161,6 +186,12 @@ class Runner_:
             else:
                 assert False, "Did not raise Failure for WatcherError!"
 
+        def does_not_apply_to_timeout_errors(self):
+            with raises(CommandTimedOut):
+                self._runner(klass=_TimingOutRunner).run(
+                    _, timeout=1, warn=True
+                )
+
     class hide:
         @trap
         def honors_config(self):
@@ -308,6 +339,36 @@ class Runner_:
             # TODO: vendor & use a color module
             assert sys.stdout.getvalue() == "\x1b[1;37mmy command\x1b[0m\n"
 
+        @trap
+        def uses_custom_format(self):
+            self._run(
+                "my command",
+                echo=True,
+                settings={"run": {"echo_format": "AA{command}ZZ"}},
+            )
+            assert sys.stdout.getvalue() == "AAmy commandZZ\n"
+
+    class dry_running:
+        @trap
+        def sets_echo_to_True(self):
+            self._run("what up", settings={"run": {"dry": True}})
+            assert "what up" in sys.stdout.getvalue()
+
+        @trap
+        def short_circuits_with_dummy_result(self):
+            runner = self._runner(run={"dry": True})
+            # Using the call to self.start() in _run_body() as a sentinel for
+            # all the work beyond it.
+            runner.start = Mock()
+            result = runner.run(_)
+            assert not runner.start.called
+            assert isinstance(result, Result)
+            assert result.command == _
+            assert result.stdout == ""
+            assert result.stderr == ""
+            assert result.exited == 0
+            assert result.pty is False
+
     class encoding:
         # NOTE: these tests just check what Runner.encoding ends up as; it's
         # difficult/impossible to mock string objects themselves to see what
@@ -435,6 +496,13 @@ class Runner_:
             assert out.getvalue() == "sup"
             assert sys.stdout.getvalue() == ""
 
+        @trap
+        def overridden_out_is_never_hidden(self):
+            out = StringIO()
+            self._runner(out="sup").run(_, out_stream=out, hide=True)
+            assert out.getvalue() == "sup"
+            assert sys.stdout.getvalue() == ""
+
         @trap
         def err_can_be_overridden(self):
             "err_stream can be overridden"
@@ -443,6 +511,13 @@ class Runner_:
             assert err.getvalue() == "sup"
             assert sys.stderr.getvalue() == ""
 
+        @trap
+        def overridden_err_is_never_hidden(self):
+            err = StringIO()
+            self._runner(err="sup").run(_, err_stream=err, hide=True)
+            assert err.getvalue() == "sup"
+            assert sys.stderr.getvalue() == ""
+
         @trap
         def pty_defaults_to_sys(self):
             self._runner(out="sup").run(_, pty=True)
@@ -530,6 +605,40 @@ class Runner_:
             assert "'oh god why'" in msg
             assert "in thread for 'handle_stdin'" in msg
 
+        def EOF_triggers_closing_of_proc_stdin(self):
+            class Fake(_Dummy):
+                pass
+
+            Fake.close_proc_stdin = Mock()
+            self._runner(klass=Fake).run(_, in_stream=StringIO("what?"))
+            Fake.close_proc_stdin.assert_called_once_with()
+
+        def EOF_does_not_close_proc_stdin_when_pty_True(self):
+            class Fake(_Dummy):
+                pass
+
+            Fake.close_proc_stdin = Mock()
+            self._runner(klass=Fake).run(
+                _, in_stream=StringIO("what?"), pty=True
+            )
+            assert not Fake.close_proc_stdin.called
+
+        @patch("invoke.runners.sys.stdin")
+        def EBADF_on_stdin_read_ignored(self, fake_stdin):
+            # Issue #659: nohup is a jerk
+            fake_stdin.read.side_effect = OSError(errno.EBADF, "Ugh")
+            # No boom == fixed
+            self._runner().run(_)
+
+        @patch("invoke.runners.sys.stdin")
+        def non_EBADF_on_stdin_read_not_ignored(self, fake_stdin):
+            # Issue #659: nohup is a jerk, inverse case
+            eio = OSError(errno.EIO, "lol")
+            fake_stdin.read.side_effect = eio
+            with raises(ThreadException) as info:
+                self._runner().run(_)
+            assert info.value.exceptions[0].value is eio
+
     class failure_handling:
         def fast_failures(self):
             with raises(UnexpectedExit):
@@ -539,6 +648,19 @@ class Runner_:
             r = self._runner(exits=17).run(_, warn=True)
             assert r.failed is True
 
+        class Failure_repr:
+            def defaults_to_just_class_and_command(self):
+                expected = "<Failure: cmd='oh hecc'>"
+                assert repr(Failure(Result(command="oh hecc"))) == expected
+
+            def subclasses_may_add_more_kv_pairs(self):
+                class TotalFailure(Failure):
+                    def _repr(self, **kwargs):
+                        return super(TotalFailure, self)._repr(mood="dejected")
+
+                expected = "<TotalFailure: cmd='onoz' mood=dejected>"
+                assert repr(TotalFailure(Result(command="onoz"))) == expected
+
         class UnexpectedExit_repr:
             def similar_to_just_the_result_repr(self):
                 try:
@@ -550,12 +672,10 @@ class Runner_:
         class UnexpectedExit_str:
             def setup(self):
                 def lines(prefix):
-                    return (
-                        "\n".join(
-                            "{} {}".format(prefix, x) for x in range(1, 26)
-                        )
-                        + "\n"
+                    prefixed = "\n".join(
+                        "{} {}".format(prefix, x) for x in range(1, 26)
                     )
+                    return prefixed + "\n"
 
                 self._stdout = lines("stdout")
                 self._stderr = lines("stderr")
@@ -577,10 +697,8 @@ Stdout: already printed
 
 Stderr: already printed
 
-""".format(
-                        _
-                    )
-                    assert str(e) == expected
+"""
+                    assert str(e) == expected.format(_)
                 else:
                     assert False, "Failed to raise UnexpectedExit!"
 
@@ -601,10 +719,8 @@ Stdout: already printed
 
 Stderr: n/a (PTYs have no stderr)
 
-""".format(
-                        _
-                    )
-                    assert str(e) == expected
+"""
+                    assert str(e) == expected.format(_)
 
             @trap
             def pty_stderr_message_wins_over_hidden_stderr(self):
@@ -660,10 +776,8 @@ stderr 23
 stderr 24
 stderr 25
 
-""".format(
-                        _
-                    )
-                    assert str(e) == expected
+"""
+                    assert str(e) == expected.format(_)
 
             @trap
             def displays_tails_of_streams_only_when_hidden(self):
@@ -1000,7 +1114,10 @@ stderr 25
                     in_stream=StringIO("foo"),
                     out_stream=StringIO(),  # null output to not pollute tests
                 )
-            assert mock_time.sleep.call_args_list == [call(0.007)] * 3
+            # Just make sure the first few sleeps all look good. Can't know
+            # exact length of list due to stdin worker hanging out til end of
+            # process. Still worth testing more than the first tho.
+            assert mock_time.sleep.call_args_list[:3] == [call(0.007)] * 3
 
     class stdin_mirroring:
         def _test_mirroring(self, expect_mirroring, **kwargs):
@@ -1214,7 +1331,7 @@ stderr 25
             runner.send_interrupt = Mock()
             try:
                 runner.run(_)
-            except:
+            except _GenericException:
                 pass
             return runner
 
@@ -1225,10 +1342,6 @@ stderr 25
             assert runner.send_interrupt.called
 
         def not_called_for_other_exceptions(self):
-            class _GenericExceptingRunner(_Dummy):
-                def wait(self):
-                    raise Exception
-
             runner = self._run_with_mocked_interrupt(_GenericExceptingRunner)
             assert not runner.send_interrupt.called
 
@@ -1240,20 +1353,185 @@ stderr 25
                 runner.run(_, pty=pty)
                 mock_stdin.assert_called_once_with(u"\x03")
 
+    class timeout:
+        def start_timer_called_with_config_value(self):
+            runner = self._runner(timeouts={"command": 7})
+            runner.start_timer = Mock()
+            assert runner.context.config.timeouts.command == 7
+            runner.run(_)
+            runner.start_timer.assert_called_once_with(7)
+
+        def run_kwarg_honored(self):
+            runner = self._runner()
+            runner.start_timer = Mock()
+            assert runner.context.config.timeouts.command is None
+            runner.run(_, timeout=3)
+            runner.start_timer.assert_called_once_with(3)
+
+        def kwarg_wins_over_config(self):
+            runner = self._runner(timeouts={"command": 7})
+            runner.start_timer = Mock()
+            assert runner.context.config.timeouts.command == 7
+            runner.run(_, timeout=3)
+            runner.start_timer.assert_called_once_with(3)
+
+        def raises_CommandTimedOut_with_timeout_info(self):
+            runner = self._runner(
+                klass=_TimingOutRunner, timeouts={"command": 7}
+            )
+            with raises(CommandTimedOut) as info:
+                runner.run(_)
+            assert info.value.timeout == 7
+            _repr = "<CommandTimedOut: cmd='nope' timeout=7>"
+            assert repr(info.value) == _repr
+            expected = """
+Command did not complete within 7 seconds!
+
+Command: 'nope'
+
+Stdout: already printed
+
+Stderr: already printed
+
+""".lstrip()
+            assert str(info.value) == expected
+
+        @patch("invoke.runners.threading.Timer")
+        def start_timer_gives_its_timer_the_kill_method(self, Timer):
+            runner = self._runner()
+            runner.start_timer(30)
+            Timer.assert_called_once_with(30, runner.kill)
+
+        def _mocked_timer(self):
+            runner = self._runner()
+            runner._timer = Mock()
+            return runner
+
+        def run_always_stops_timer(self):
+            runner = _GenericExceptingRunner(Context())
+            runner.stop_timer = Mock()
+            with raises(_GenericException):
+                runner.run(_)
+            runner.stop_timer.assert_called_once_with()
+
+        def stop_timer_cancels_timer(self):
+            runner = self._mocked_timer()
+            runner.stop_timer()
+            runner._timer.cancel.assert_called_once_with()
+
+        def timer_aliveness_is_test_of_timing_out(self):
+            # Might be redundant, but easy enough to unit test
+            runner = Runner(Context())
+            runner._timer = Mock()
+            runner._timer.is_alive.return_value = False
+            assert runner.timed_out
+            runner._timer.is_alive.return_value = True
+            assert not runner.timed_out
+
+        def timeout_specified_but_no_timer_means_no_exception(self):
+            # Weird corner case but worth testing
+            runner = Runner(Context())
+            runner._timer = None
+            assert not runner.timed_out
+
     class stop:
         def always_runs_no_matter_what(self):
-            class _ExceptingRunner(_Dummy):
-                def wait(self):
-                    raise OhNoz()
-
-            runner = _ExceptingRunner(context=Context())
+            runner = _GenericExceptingRunner(context=Context())
             runner.stop = Mock()
-            try:
+            with raises(_GenericException):
                 runner.run(_)
-            except OhNoz:
-                runner.stop.assert_called_once_with()
-            else:
-                assert False, "_ExceptingRunner did not except!"
+            runner.stop.assert_called_once_with()
+
+    class asynchronous:
+        def returns_Promise_immediately_and_finishes_on_join(self):
+            # Dummy subclass with controllable process_is_finished flag
+            class _Finisher(_Dummy):
+                _finished = False
+
+                @property
+                def process_is_finished(self):
+                    return self._finished
+
+            runner = _Finisher(Context())
+            # Set up mocks and go
+            runner.start = Mock()
+            for method in self._stop_methods:
+                setattr(runner, method, Mock())
+            result = runner.run(_, asynchronous=True)
+            # Got a Promise (its attrs etc are in its own test subsuite)
+            assert isinstance(result, Promise)
+            # Started, but did not stop (as would've happened for disown)
+            assert runner.start.called
+            for method in self._stop_methods:
+                assert not getattr(runner, method).called
+            # Set proc completion flag to truthy and join()
+            runner._finished = True
+            result.join()
+            for method in self._stop_methods:
+                assert getattr(runner, method).called
+
+        @trap
+        def hides_output(self):
+            # Run w/ faux subproc stdout/err data, but async
+            self._runner(out="foo", err="bar").run(_, asynchronous=True).join()
+            # Expect that default out/err streams did not get printed to.
+            assert sys.stdout.getvalue() == ""
+            assert sys.stderr.getvalue() == ""
+
+        def does_not_forward_stdin(self):
+            class MockedHandleStdin(_Dummy):
+                pass
+
+            MockedHandleStdin.handle_stdin = Mock()
+            runner = self._runner(klass=MockedHandleStdin)
+            runner.run(_, asynchronous=True).join()
+            # As with the main test for setting this to False, we know that
+            # when stdin is disabled, the handler is never even called (no
+            # thread is created for it).
+            assert not MockedHandleStdin.handle_stdin.called
+
+        def leaves_overridden_streams_alone(self):
+            # NOTE: technically a duplicate test of the generic tests for #637
+            # re: intersect of hide and overridden streams. But that's an
+            # implementation detail so this is still valuable.
+            klass = self._mock_stdin_writer()
+            out, err, in_ = StringIO(), StringIO(), StringIO("hallo")
+            runner = self._runner(out="foo", err="bar", klass=klass)
+            runner.run(
+                _,
+                asynchronous=True,
+                out_stream=out,
+                err_stream=err,
+                in_stream=in_,
+            ).join()
+            assert out.getvalue() == "foo"
+            assert err.getvalue() == "bar"
+            assert klass.write_proc_stdin.called  # lazy
+
+    class disown:
+        @patch.object(threading.Thread, "start")
+        def starts_and_returns_None_but_does_nothing_else(self, thread_start):
+            runner = Runner(Context())
+            runner.start = Mock()
+            not_called = self._stop_methods + ["wait"]
+            for method in not_called:
+                setattr(runner, method, Mock())
+            result = runner.run(_, disown=True)
+            # No Result object!
+            assert result is None
+            # Subprocess kicked off
+            assert runner.start.called
+            # No timer or IO threads started
+            assert not thread_start.called
+            # No wait or shutdown related Runner methods called
+            for method in not_called:
+                assert not getattr(runner, method).called
+
+        def cannot_be_given_alongside_asynchronous(self):
+            with raises(ValueError) as info:
+                self._runner().run(_, asynchronous=True, disown=True)
+            sentinel = "Cannot give both 'asynchronous' and 'disown'"
+            assert sentinel in str(info.value)
 
 
 class _FastLocal(Local):
@@ -1345,6 +1623,12 @@ class Local_:
                 assert e.type == OSError
                 assert str(e.value) == "wat"
 
+        @mock_pty(os_close_error=True)
+        def stop_mutes_errors_on_pty_close(self):
+            # Another doesn't-blow-up test, this time around os.close() of the
+            # pty itself (due to os_close_error=True)
+            self._run(_, pty=True)
+
         class fallback:
             @mock_pty(isatty=False)
             def can_be_overridden_by_kwarg(self):
@@ -1411,6 +1695,37 @@ class Local_:
             env = mock_os.execve.call_args_list[0][0][2]
             assert env == expected
 
+    class close_proc_stdin:
+        def raises_SubprocessPipeError_when_pty_in_use(self):
+            with raises(SubprocessPipeError):
+                runner = Local(Context())
+                runner.using_pty = True
+                runner.close_proc_stdin()
+
+        def closes_process_stdin(self):
+            runner = Local(Context())
+            runner.process = Mock()
+            runner.using_pty = False
+            runner.close_proc_stdin()
+            runner.process.stdin.close.assert_called_once_with()
+
+    class timeout:
+        @patch("invoke.runners.os")
+        def kill_uses_self_pid_when_pty(self, mock_os):
+            runner = self._runner()
+            runner.using_pty = True
+            runner.pid = 50
+            runner.kill()
+            mock_os.kill.assert_called_once_with(50, signal.SIGKILL)
+
+        @patch("invoke.runners.os")
+        def kill_uses_self_process_pid_when_not_pty(self, mock_os):
+            runner = self._runner()
+            runner.using_pty = False
+            runner.process = Mock(pid=30)
+            runner.kill()
+            mock_os.kill.assert_called_once_with(30, signal.SIGKILL)
+
 
 class Result_:
     def nothing_is_required(self):
@@ -1425,6 +1740,9 @@ class Result_:
     def shell_defaults_to_empty_string(self):
         assert Result().shell == ""
 
+    def encoding_defaults_to_local_default_encoding(self):
+        assert Result().encoding == default_encoding()
+
     def env_defaults_to_empty_dict(self):
         assert Result().env == {}
 
@@ -1442,3 +1760,104 @@ class Result_:
 
     def repr_contains_useful_info(self):
         assert repr(Result(command="foo")) == "<Result cmd='foo' exited=0>"
+
+    class tail:
+        def setup(self):
+            self.sample = "\n".join(str(x) for x in range(25))
+
+        def returns_last_10_lines_of_given_stream_plus_whitespace(self):
+            expected = """
+
+15
+16
+17
+18
+19
+20
+21
+22
+23
+24"""
+            assert Result(stdout=self.sample).tail("stdout") == expected
+
+        def line_count_is_configurable(self):
+            expected = """
+
+23
+24"""
+            tail = Result(stdout=self.sample).tail("stdout", count=2)
+            assert tail == expected
+
+        def works_for_stderr_too(self):
+            # Dumb test is dumb, but whatever
+            expected = """
+
+23
+24"""
+            tail = Result(stderr=self.sample).tail("stderr", count=2)
+            assert tail == expected
+
+        @patch("invoke.runners.encode_output")
+        def encodes_with_result_encoding(self, encode):
+            Result(stdout="foo", encoding="utf-16").tail("stdout")
+            encode.assert_called_once_with("\n\nfoo", "utf-16")
+
+
+class Promise_:
+    def exposes_read_only_run_params(self):
+        runner = _runner()
+        promise = runner.run(
+            _, pty=True, encoding="utf-17", shell="sea", asynchronous=True
+        )
+        assert promise.command == _
+        assert promise.pty is True
+        assert promise.encoding == "utf-17"
+        assert promise.shell == "sea"
+        assert not hasattr(promise, "stdout")
+        assert not hasattr(promise, "stderr")
+
+    class join:
+        # NOTE: high level Runner lifecycle mechanics of join() (re: wait(),
+        # process_is_finished() etc) are tested in main suite.
+
+        def returns_Result_on_success(self):
+            result = _runner().run(_, asynchronous=True).join()
+            assert isinstance(result, Result)
+            # Sanity
+            assert result.command == _
+            assert result.exited == 0
+
+        def raises_main_thread_exception_on_kaboom(self):
+            runner = _runner(klass=_GenericExceptingRunner)
+            with raises(_GenericException):
+                runner.run(_, asynchronous=True).join()
+
+        def raises_subthread_exception_on_their_kaboom(self):
+            class Kaboom(_Dummy):
+                def handle_stdout(self, **kwargs):
+                    raise OhNoz()
+
+            runner = _runner(klass=Kaboom)
+            promise = runner.run(_, asynchronous=True)
+            with raises(ThreadException) as info:
+                promise.join()
+            assert isinstance(info.value.exceptions[0].value, OhNoz)
+
+        def raises_Failure_on_failure(self):
+            runner = _runner(exits=1)
+            promise = runner.run(_, asynchronous=True)
+            with raises(Failure):
+                promise.join()
+
+    class context_manager:
+        def calls_join_or_wait_on_close_of_block(self):
+            promise = _runner().run(_, asynchronous=True)
+            promise.join = Mock()
+            with promise:
+                pass
+            promise.join.assert_called_once_with()
+
+        def yields_self(self):
+            promise = _runner().run(_, asynchronous=True)
+            with promise as value:
+                assert value is promise
